#%%%%%%%%%%%%%%%%%%%%%%%  Prepare for testing   %%%%%%%%%%%%%%
import os
import backtest_pkg.backtest_portfolio as bt
import pandas as pd 
from IPython.display import display
import importlib

os.chdir(r'M:\Share\Colleagues\Andy\Python Project\Backtest Module')
price_data = pd.read_csv('pkg_test/Adjusted_Price.csv', index_col=0, parse_dates=True)

#########################################################################
########                 Portfolio construction           ###############
#########################################################################
#%%%%%%%%%%%%%%%%%%%%%%%  Portfolio from weight   %%%%%%%%%%%%%%%
importlib.reload(bt)

# Small testing date:
small_price_data = price_data.iloc[:10, :5]
small_weight = pd.DataFrame(data=[[1, 2, 3]], index=[price_data.index[0]], columns=price_data.columns[:3])
small_share = pd.DataFrame(data=1, index=[price_data.index[0]], columns=price_data.columns[:3])

# Initiate a portfolio by weight:
small_port = bt.portfolio(small_weight, name='Small Portfolio')
# Better way to initialize a portfolio:
# small_port = bt.portfolio(weight = small_weight, name='Small Portfolio')
small_port.set_price(small_price_data)
display(small_port.weight)

##### Check some properties:
# End date:
small_port.end_date = None   # Default: last date of price data
print(f'End date (default): {small_port.end_date:%Y-%m-%d}')
small_port.end_date = pd.datetime(2013, 1, 10)  # Manually set
print(f'End date (set): {small_port.end_date:%Y-%m-%d}')

# Trading_status:
small_port.trading_status = None   # Default: TRUE if price is available
display(f'Trading Status (default):')
display(small_port.trading_status)
    
trading_status = small_price_data.notna()
trading_status.loc[:'2013-01-10', 'ALB SQ Equity'] = False
small_port.trading_status = trading_status  # Set for specific requirement
display(f'Trading Status (set):')
display(small_port.trading_status)

# Daily return: calculate from price, the first available date has return 1.
display(f'Daily Return:')
display(small_port.daily_ret)

#%%%%%%%%%%%%%%%%%%   Portfolio from share   %%%%%%%%%%%%%%%%%%%%%
# Initiate a portfolio by share:
small_port = bt.portfolio(share = small_share, name='Small share Portfolio')
small_port.set_price(small_price_data)
display(small_port.weight)

##### Check some properties:
# End date:
small_port.end_date = None   # Default: last date of price data
print(f'End date (default): {small_port.end_date:%Y-%m-%d}')
small_port.end_date = pd.datetime(2013, 1, 10)  # Manually set
print(f'End date (set): {small_port.end_date:%Y-%m-%d}')

# Trading_status:
small_port.trading_status = None   # Default: TRUE if price is available
display(f'Trading Status (default):')
display(small_port.trading_status)
    
trading_status = small_price_data.notna()
trading_status.loc[:'2013-01-10', 'ALB SQ Equity'] = False
small_port.trading_status = trading_status  # Set for specific requirement
display(f'Trading Status (set):')
display(small_port.trading_status)

# Daily return: calculate from price, the first available date has return 1.
display(f'Daily Return:')
display(small_port.daily_ret)

#####################    Error handling   ##############################
importlib.reload(bt)
#%%%%%%%%%%%%%%%%%%%%   Weight in untradable security  %%%%%%%%%%%%%%%%%
small_weight = pd.DataFrame(data=[[1, 2, 3]], index=[small_price_data.index[0]], columns= small_price_data.columns[-3:])
small_share = pd.DataFrame(data=1, index=[small_price_data.index[0]], columns=small_price_data.columns[-3:])

small_port = bt.portfolio(small_weight, name='Small error Portfolio')
small_port.set_price(small_price_data)
display(small_weight)
display(small_port.weight)

#%%%%%%%%%%%%%%%%%%%   Share in untradable security   %%%%%%%%%%%%%%%%%%%%%
small_port = bt.portfolio(share=small_share, name='Small error Portfolio from share')
small_port.set_price(small_price_data)
display(small_share)
display(small_port.weight)

#%%%%%%%%%%%%%%%%%%%%%%  Unknown tickers   %%%%%%%%%%%%%%%%%
small_weight = pd.DataFrame(data=[[1, 2, 3]], index=[small_price_data.index[0]], columns= small_price_data.columns[:3])
small_weight['Strange Equity']=1
small_port = bt.portfolio(small_weight, name='Small Portfolio with unknown ticker')
small_port.set_price(small_price_data)
display('Input weight:')
display(small_weight)
display('Output weight:')
display(small_port.weight)
#%%%%%%%%%%%%%%%%%%%%%%%    Outrange date   %%%%%%%%%%%%%%%%%%%%%%%%
small_weight = pd.DataFrame(data=[[1, 2, 3]], index=[small_price_data.index[0]], columns= small_price_data.columns[:3])
small_weight.loc[pd.to_datetime('2019-01-01'), :] = 1
small_port = bt.portfolio(small_weight, name='Small Portfolio with unknown ticker')
small_port.set_price(small_price_data)
display('Input weight:')
display(small_weight)
display('Final weight:')
display(small_port.weight)
#%%%%%%%%%%%%%%%%%%%%%%  Unknown tickers in share   %%%%%%%%%%%%%%%%%
small_share = pd.DataFrame(data=1, index=[small_price_data.index[0]], columns= small_price_data.columns[:3])
small_share['Strange Equity']=1
small_port = bt.portfolio(share = small_share, name='Small Portfolio with unknown ticker')
small_port.set_price(small_price_data)
display('Input weight:')
display(small_share)
display('Final weight:')
display(small_port.weight)
#%%%%%%%%%%%%%%%%%%%%%%  Outrange date in share   %%%%%%%%%%%%%%%%%
small_share = pd.DataFrame(data=1, index=[small_price_data.index[0]], columns= small_price_data.columns[:3])
small_share.loc[pd.to_datetime('2019-01-01'), :]=1
small_port = bt.portfolio(share = small_share, name='Small Portfolio with unknown ticker')
small_port.set_price(small_price_data)
display('Input weight:')
display(small_share)
display('Final weight:')
display(small_port.weight)

########################################################################
##################         Backtesting                   ###############
########################################################################

#%%%%%%%%%%%%%%   Construct a small portfolio   %%%%%%%%%%%%%%
importlib.reload(bt)
small_price_data = price_data.iloc[:10, :5]

# Initiate an equal weight portfolio by weight:
small_weight = pd.DataFrame(data=1, index=[small_price_data.index[0]], columns=small_price_data.columns)
small_port = bt.portfolio(small_weight, name='Small Portfolio')
small_port.set_price(small_price_data)

#%%%%%%%%%%%%%%%%%%%%%   Simple drift   %%%%%%%%%%%%%%%%%%%%
drifting_weight = small_port._drift_weight(small_port.weight)
display('Simple drift')
display(drifting_weight)

#%%%%%%%%%%%%%%%%%%%%    Rebalanced drift   %%%%%%%%%%%%%%%%%
rebalance_date = pd.datetime(2013, 1, 10)
rebalanced_weight = pd.DataFrame(data=1, index=[rebalance_date], columns=small_price_data.columns)
drifting_weight_rebalanced = drifting_weight.copy()
new_weights = small_port._drift_weight(drifting_weight.loc[[rebalance_date], :], rebalanced_weight=rebalanced_weight)
drifting_weight_rebalanced.loc[new_weights.index, :] = new_weights

display(f'Rebalance on {rebalance_date: %y-%m-%d} to equal weight')
display(drifting_weight_rebalanced)

#%%%%%%%%%%%%%%%%%%%   Rebalance drift with untradable stocks  %%%%%%%%%%%%%%%
trading_status = small_price_data.notna()
trading_status.loc[rebalance_date, :] = [False]*2 + [True]*3
small_port.trading_status = trading_status
drifting_weight_rebalanced = drifting_weight.copy()
new_weights = small_port._drift_weight(drifting_weight.loc[[rebalance_date], :], rebalanced_weight=rebalanced_weight)
drifting_weight_rebalanced.loc[new_weights.index, :] = new_weights

display(f'Rebalance on {rebalance_date: %y-%m-%d} to equal weight (untradable on SZG GY, KER FP)')
display(drifting_weight_rebalanced)

#%%%%%%%%%%%%%%%%%%%%    Extend weight %%%%%%%%%%%%%%%%%%%%%%%
importlib.reload(bt)
small_price_data = price_data.iloc[:10, :5]

# Initiate an equal weight portfolio by weight:
small_weight = pd.DataFrame(data=1, index=small_price_data.index[0::5], columns=small_price_data.columns)
small_port = bt.portfolio(small_weight, name='Small Portfolio')
small_port.set_price(small_price_data)

display('Portfolio weight:')
display(small_port.weight)
display('Extended portfolio weight:')
display(small_port.ex_weight)

#%%%%%%%%%%%%%%%%%%%%%%%%   Daily return and total return    %%%%%%%%%%%%%%%%%
display('Daily return:')
display(small_port.port_daily_ret)
display('Total return:')
display(small_port.port_total_ret)
display(small_port.port_total_ret.plot())

#%%%%%%%%%%%%%%%%%%%%%%   Extend weight with untradable   %%%%%%%%%%%%%%
# Prepare weight:
small_weight = pd.DataFrame(data=1, index=small_price_data.index[0::5], columns=small_price_data.columns)
# Prepare trading status:
rebalance_date = small_price_data.index[5]
trading_status = small_price_data.notna()
trading_status.loc[rebalance_date, :] = [False]*2 + [True]*3
# Construct portfolio
small_port = bt.portfolio(small_weight, name='Small Portfolio')
small_port.set_price(small_price_data, trading_status=trading_status)

display('Portfolio weight:')
display(small_port.weight)
display('Extended portfolio weight:')
display(small_port.ex_weight)

#%%%%%%%%%%%%%%%%%%%%%%%%%   Daily and total return   %%%%%%%%%%%%%%%%%%
display('Daily return:')
display(small_port.port_daily_ret)
display('Total return:')
display(small_port.port_total_ret)
display(small_port.port_total_ret.plot())

#%%%%%%%%%%%%%%%%%%    Simple backtest (no benchmark)   %%%%%%%%%%%%%%%%%
importlib.reload(bt)
port_weight = pd.DataFrame(data=0, index=small_price_data.index[[0, 5]], columns=small_price_data.columns) 
port_weight.iloc[0, [0, 1]] = 1
port_weight.iloc[1, [2, 3]] = 1
# Construct portfolio
small_port = bt.portfolio(weight=port_weight, name='Selection Portfolio')
small_port.set_price(small_price_data)

small_port.backtest(plot=True)

#%%%%%%%%%%%%%%%%%%    Backtest with benchmark   %%%%%%%%%%%%%%%%%%%%%%%
# Prepare weight:
benchmark_weight = pd.DataFrame(data=1, index=small_price_data.index[[0, 5]], columns=small_price_data.columns)
port_weight = pd.DataFrame(data=0, index=small_price_data.index[[0, 5]], columns=small_price_data.columns) 
port_weight.iloc[0, [0, 1]] = 1
port_weight.iloc[1, [2, 3]] = 1
# Construct portfolio
small_port = bt.portfolio(
    weight=port_weight, benchmark=benchmark_weight,
    name='Selection Portfolio', benchmark_name='Equal weight Benchmark')
small_port.set_price(small_price_data)

display(small_port.backtest())
display(small_port.performance_plot())

#%%%%%%%%%%%%%%%%%%    Backtest with benchmark (Large portfolio) %%%%%%%%%%%%%%%%%%%%%%%
seed = 56

# Prepare end date, weights for portfolio and benchmark:
testing_length = 300
end_date = price_data.index[testing_length]
benchmark_weight = pd.DataFrame(data=1, index=price_data.index[0:testing_length:20], columns=price_data.columns)
port_weight = pd.DataFrame(data=0, index=price_data.index[0:testing_length:20], columns=price_data.columns)
for i in range(port_weight.shape[0]):
    stock_selection = port_weight.columns.to_series().sample(100, random_state = seed*(i+1))
    port_weight.iloc[i, :][stock_selection] = 1
# Construct portfolio
large_port= bt.portfolio(
    weight=port_weight, benchmark=benchmark_weight, end_date=end_date,
    name='Selection Portfolio', benchmark_name='Equal weight Benchmark')
large_port.set_price(price_data)

display(large_port.backtest())
display(large_port.performance_plot())
display(large_port.performance_summary())



###########################################################################
#################         Performance                 #####################
###########################################################################
#%%%%%%%%%%%%%%%%%%%%%%%
small_port.performance_plot()
small_port.performance_summary()















































#%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%
period_ts = pd.Series(bt_result.index.map(lambda s: (s>weight.index).sum()+(s>end_date)), index=bt_result.index)
period_ts.name = 'Period'
# bt_result['Active Return'].groupby(period_ts).agg({'Active Return':lambda s:s.cumprod().tail(1),'Tracking Error': 'std'})

port_ret= select_port.port_daily_ret
bm_ret = select_port.benchmark.port_daily_ret
daily_active_ret = port_ret - bm_ret
port_period_ret = port_ret.groupby(period_ts).agg(Portfolio_Return = 'prod') -1
bm_period_ret = bm_ret.groupby(period_ts).agg(Benchmark_Return = 'prod') -1
agg_df = pd.concat([port_period_ret, bm_period_ret], axis=1)
agg_df['Active_Return'] = agg_df.iloc[:, 0] - agg_df.iloc[:, 1]
agg_df['Tracking_Error'] = daily_active_ret.groupby(period_ts).agg('std')
agg_df['Sharpe_Ratio'] = agg_df.Active_Return/agg_df.Tracking_Error
agg_df = agg_df.drop(index=0)
agg_df

#%%



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt 
import warnings
from math import sqrt



def get_price_from_BB(tickers, start_date, end_date):
    'TBA'
    '''
    Download price data from Bloomberg. Ensusre a bloomberg connection is valid and the library pdblp is installed.
    tickers: a list of Bloomberg tickers in the universe.
    start_date: start date of the price data
    end_date: end date of the price data
    '''
    pass

def annualized_performance_metric(daily_ret_ts, tolerance=10**(-4), annual_trading_days=250):
    output = pd.Series()
    annualized_return= daily_ret_ts.prod()**(annual_trading_days/len(daily_ret_ts))-1
    output['Annualized_Return'] =annualized_return
    annualized_volatility = daily_ret_ts.std()*sqrt(annual_trading_days)
    if annualized_volatility < tolerance:
        annualized_volatility=np.nan
    output['Annualized_Volatility'] = annualized_volatility
    sharpe_ratio = annualized_return/annualized_volatility
    output['Annualized_Sharpe_Ratio'] = sharpe_ratio

    return output

def period_performance_metric(daily_ret_ts, tolerance=10**(-4)):
    output = pd.Series()
    period_return= daily_ret_ts.prod()-1
    output['Period_Return'] =period_return
    period_volatility = daily_ret_ts.std()*sqrt(len(daily_ret_ts))
    if period_volatility < tolerance:
        period_volatility=np.nan
    output['Period_Volatility'] = period_volatility
    sharpe_ratio = period_return/period_volatility
    output['Sharpe_Ratio'] = sharpe_ratio

    return output


def active_performance_metric(port_ret_ts, bm_ret_ts, tolerance=10**(-4)):
    assert (port_ret_ts.index == bm_ret_ts.index).all(), 'Two time series should be in the same period!'
    output = pd.Series()
    output['Active_Return'] = port_ret_ts.prod() - bm_ret_ts.prod()
    active_risk = (port_ret_ts-bm_ret_ts).std()*sqrt(len(port_ret_ts)) 
    if active_risk < tolerance:
        active_risk = np.nan
    output['Active_Risk'] = active_risk
    output['Information_Ratio'] = output['Active_Return']/output['Active_Risk']

    return output

def max_drawdown(total_ret_ts):
    previous_peak = total_ret_ts.cummax()
    drawdown_ts = (previous_peak - total_ret_ts)/previous_peak 
    return max(drawdown_ts)

class portfolio:
    '''
    The universe and the valid testing period will be defined by the price data.
    '''
    def __init__(self, weight=None, share=None, benchmark=None, end_date=None, name='Portfolio', benchmark_name='Benchmark'):
        '''
        weight: a df with row-names date, col-name security id, value the portfolio weight (not necessarily normalized) of col-security at row-date. 
        share: a df with row-names date, col-name security id, value the portfolio shares of col-security at row date. 
        benchmark: a df with row-names date, col-name security id, value the benchmark weight. 
        end_date: date to end backtest 
        name: the name of the portfolio
        '''
        self._weight = weight
        self._weight_new = weight is not None
        self.share = share
        self.share_new = share is not None
        # Construct a portfolio object if benchmark is given by weights:
        if isinstance(benchmark, pd.DataFrame):
            self.benchmark = portfolio(weight=benchmark, end_date=end_date, name=benchmark_name)
        elif isinstance(benchmark, portfolio) or (benchmark is None):
            self.benchmark = benchmark
        else:
            warnings.warn('Unknown benchmark type!')
            self.benchmark = None
        self._end_date = end_date
        self.name = name
        self.benchmark_name = benchmark_name

    def _adjust(self, df):
        assert self.__price is not None, "No price data!"
        # Adjust index(dates) withing price.index
        out_range_date = df.index.difference(self.__price.index)
        if len(out_range_date)>0:
            print(f'Skipping outrange dates:\n{out_range_date.values}')
            df = df.loc[df.index & self.__price.index, :]
        # Adjust columns(tickers) withing price.columns, 
        unknown_ticker = df.columns.difference(self.__price.columns)
        if len(unknown_ticker)>0:
            print(f'Removing unkown tickers:\n{unknown_ticker.values}')
            df = df.loc[:, df.columns & self.__price.columns]
        return df

    @property
    def weight(self):
        '''
        Lazy calculate _weight given share and __price.
        '''
        # Must set price to make the weight available.
        assert self.__price is not None, 'No price data!'

        if self._weight is not None:
            if self._weight_new:
                self._weight = self._adjust(self._weight)
                # Mask weights on available dates:
                self._weight = self._weight.where(self.trading_status, other = 0)
                # Normalize rows before return:
                self._weight = self._weight.divide(self._weight.sum(axis=1), axis=0).fillna(0)
                self._weight_new = False
        else:
            # Load price and share data to derive weights:
            assert self.share is not None, 'No weight and no share date!'
            if self.share_new:
                self.share = self._adjust(self.share)
                self.share_new = False
            price_data = self.__price.copy().loc[self.share.index, self.share.columns]

            # Construct the weights:
            self._weight = self.share * price_data
            self._weight_new = False
            # Mask weights on available dates:
            self._weight = self._weight.where(self.trading_status, other = 0)
            # Normalize rows before return:
            self._weight = self._weight.divide(self._weight.sum(axis=1), axis=0).fillna(0)

        return self._weight
    @weight.setter 
    def weight(self, weight_df):
        self._weight = weight_df
        self._weight_new = True
        try:
            del self._ex_weight
            del self.port_daily_ret
            del self.port_total_ret
        except AttributeError:
            pass

    @property
    def end_date(self):
        if self._end_date is None:
            self._end_date = max(self.__price.index)
        return self._end_date
    @end_date.setter 
    def end_date(self, value):
        self._end_date = value

    ######################    Price and related attributes   ########################
    def set_price(self, price_data, trading_status=None):
        '''
        price_data: a df with row-names date, col-name security id, value the price of col-security at row date. 
        trading_status: a df with row-names date, col-name security id, boolean value indicate if col-security is tradable at row-date. 
        '''
        # Price for backtesting is private attribute.
        self.__price = price_data
        self._trading_status = trading_status
        if self.benchmark:
            self.benchmark.__price = price_data
            self.benchmark._trading_status = trading_status

    @property
    def daily_ret(self):
        '''
        Lazy calculate _daily_ret from __price attribute.
        '''
        try:
            return self._daily_ret
        except AttributeError:
            self._daily_ret = self.__price.ffill()/self.__price.ffill().shift(1).bfill(limit=1)
            # self._daily_ret.iloc[0, :] = 1
            return self._daily_ret

    @property
    def trading_status(self): 
        '''
        Lazy calcuate _trading_status from __price.
        '''
        if self._trading_status is None:
            self._trading_status = self.__price.notnull() # Valid for trade only if price exists 
        return self._trading_status
    @trading_status.setter
    def trading_status(self, value):
        self._trading_status = value

    #####################  Backtesting methods   ####################
    def _drift_weight(self, initial_weight, rebalanced_weight=None, end=None):
        '''
        initial_weight: weight before rebalance with shape (1, n)
        rebalanced_weight: weight after rebalance with shape (1, n), same index as initial weight.
        end: end date of the drifting period.
        '''
        # Prepare end of drifting period:
        if end is None:
            end = self.end_date
        elif end > self.end_date:
            print(f'Invalid end date, set to {self.end_date} (portfolio end date)!')
            end = self.end_date

        ######################    Rebalance    ########################
        # Prepare the initial and rebalanced weight:
        if initial_weight.shape[0]>1:
            print('Only the first initial weight will be used')
            initial_weight = initial_weight.iloc[[0], :]
        initial_weight = initial_weight/initial_weight.iloc[0, :].sum()

        if rebalanced_weight is None:
            rebalanced_weight = initial_weight
        else:
            if rebalanced_weight.shape[0]>1:
                print('Only the first rebalance weight will be used!')
                rebalanced_weight = rebalanced_weight.iloc[[0], :]
            rebalanced_weight = rebalanced_weight/rebalanced_weight.iloc[0, :].sum()
            
            assert initial_weight.index[0] == rebalanced_weight.index[0], 'Inconsistent weight data!'

            # Determine tradable tickers from self.trading_status:
            rebalanced_date = initial_weight.index[0]
            trading_status = self.trading_status.loc[[rebalanced_date], :]

            # Two weight vectors will be calcuate: one for rebalance, one for roll forward
            rebalanced_weight = rebalanced_weight.where(trading_status, other=0)
            roll_forward_weight = initial_weight.where(~trading_status, other=0)
            roll_forward_total = roll_forward_weight.iloc[0, :].sum()
            if roll_forward_total<1:
                rebalanced_total = rebalanced_weight.iloc[0, :].sum()
                adjustment_factor = (1-roll_forward_total)/rebalanced_total
                rebalanced_weight = rebalanced_weight*adjustment_factor
                rebalanced_weight = rebalanced_weight+roll_forward_weight
            else:
                rebalanced_weight = roll_forward_weight
            assert abs(rebalanced_weight.iloc[0, :].sum()-1)<1e-4, 'Abnormal rebalanced weight!'


        ########################    Drifting   ##################
        # Prepare period price data:
        period_index = self.__price.index
        period_index = period_index[(period_index>=initial_weight.index[0]) & (period_index<=end)]
        period_price = self.__price.loc[period_index, :].ffill()

        # Total returns:
        total_return = period_price/period_price.iloc[0,:]
        # Drifting weights:
        drift_weight = rebalanced_weight.reindex(period_index).ffill()
        drift_weight = drift_weight * total_return 
        drift_weight = drift_weight.div(drift_weight.sum(axis=1), axis=0).fillna(0)

        return drift_weight

    @property
    def ex_weight(self):
        '''
        Extend the weight to all dates before self.end_date.
        weight: The weight to extend, represented a df with row-names date, col-name security id, value the portfolio weight of col-security at row-date. 
        '''
        try:
            return self._ex_weight
        except AttributeError:
            # Prepare the index after extention: (From first weight to end date)
            extend_period = self.__price.index
            extend_period = extend_period[(extend_period>=self.weight.index[0])&(extend_period<=self.end_date)]
            extend_weight = self.weight.reindex(extend_period)

            # Prepare the tuples for start and end date in each rebalancing period:
            rebalance_dates = pd.Series(self.weight.index)
            rebalance_start_end = zip(rebalance_dates,rebalance_dates.shift(-1, fill_value= self.end_date))

            # Initial holdings are all 0:
            initial_weight = pd.DataFrame(0, index=[extend_period[0]], columns=self.__price.columns)

            # Loop over each rebalancing period:
            for start, end in rebalance_start_end:
                rebalanced_weight = self.weight.loc[[start], :]
                period_weight = self._drift_weight(initial_weight=initial_weight,rebalanced_weight=rebalanced_weight, end=end) 
                extend_weight.loc[start:end, :] =  period_weight
                initial_weight = extend_weight.loc[[end], :]

            self._ex_weight = extend_weight
            return self._ex_weight

    @property
    def port_daily_ret(self):
        try:
            return self._port_daily_ret
        except AttributeError:
            daily_ret = self.daily_ret.copy()
            ex_weight = self.ex_weight
            daily_ret = daily_ret.loc[daily_ret.index&ex_weight.index, daily_ret.columns&ex_weight.columns]
            # Calculate portfolio daily return: 
            port_daily_ret = (ex_weight.shift(1)*daily_ret).sum(axis = 1)
            port_daily_ret[0] = 1
            self._port_daiy_ret = port_daily_ret
            return port_daily_ret
        
    @property 
    def port_total_ret(self):
        try:
            return self._port_total_ret
        except AttributeError:
            self._port_total_ret = self.port_daily_ret.cumprod()
            return self._port_total_ret

    def backtest(self, plot=False):
        '''
        Backtest portfolio performance over given period.
        '''
        # Setup price and trading status:
        port_total_ret_df = self.port_total_ret.to_frame(name=self.name)-1
        if self.benchmark is not None:
            bm_total_ret_df = self.benchmark.port_total_ret.to_frame(name=self.benchmark.name)-1
        else:
            bm_total_ret_df = pd.DataFrame(0, index=port_total_ret_df.index, columns=['Empty Portfolio'])

        result = pd.concat([port_total_ret_df, bm_total_ret_df], axis=1, sort=False)
        result['Active Weight'] = result.iloc[:,0] - result.iloc[:,1]
        self.backtest_result = result
        
        if plot:
            self.performance_plot()

        return self.backtest_result
        

    ####################    Performance     ##############################
    def performance_plot(self):
        '''
        Plot 2 figures:
        1. The portfolio return and benchmark return over backtest period.
        2. The active return over the backtest period.
        '''
        result = self.backtest_result
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize = (7, 10))
        # make a little extra space between the subplots
        fig.subplots_adjust(hspace=0.5)

        # Upper figure for total return:
        ax1.plot(result.iloc[:, 0], label=result.columns[0])
        ax1.plot(result.iloc[:, 1], label=result.columns[1])
        ax1.tick_params(axis='x', rotation=25)
        ax1.grid(color='grey', ls='--')
        ax1.legend()
        ax1.set_title('Total Return')
        # Lower figure for active return:
        ax2.plot(result.iloc[:, 2])
        ax2.tick_params(axis='x', rotation=25)
        ax2.grid(color='grey', ls='--')
        ax2.set_title('Active Return')

        plt.show() 
        # return fig    
    
    def performance_summary(self):
        '''
        Provide a table of total return, volitility, Sharpe ratio for portfoilo, benchmark and active weight.
        '''
        try:
            return self._performance_summary
        except AttributeError:
            summary_table=pd.DataFrame(columns=['Return', 'Volatility', 'Sharpe_Ratio'])
            port_metric = period_performance_metric(self.port_daily_ret)
            port_metric.index = summary_table.columns
            summary_table.loc[self.name, :] = port_metric
            bm_metric = period_performance_metric(self.benchmark.port_daily_ret) 
            bm_metric.index = summary_table.columns
            summary_table.loc[self.benchmark.name, :] = bm_metric
            active_metric = active_performance_metric(self.port_daily_ret, self.benchmark.port_daily_ret) 
            active_metric.index = summary_table.columns
            summary_table.loc['Active', :] = active_metric
            # Formatting before output:
            summary_table = summary_table.style.format({
                'Return': '{:,.2%}'.format,
                'Volatility': '{:,.2%}'.format,
                'Sharpe_Ratio': '{:,.2f}'.format,
            })

            self._performance_summary = summary_table
            return self._performance_summary

    @property
    def period_performance(self):
        try:
            return self._period_performance
        except AttributeError:
            # Prepare portfolio, benchmark, active return:
            port_ret= self.port_daily_ret
            bm_ret = self.benchmark.port_daily_ret
            daily_active_ret = port_ret - bm_ret
            # Label each period by rebalance date from weight attribute:
            period_ts = pd.Series(port_ret.index.map(lambda s: (s>self.weight.index).sum()), index=port_ret.index)
            period_ts.name = 'Period'
            # Calculate performance metric on each period:
            period_result = pd.DataFrame()
            period_result[self.name] = port_ret.groupby(period_ts).agg('prod') -1
            period_result[self.benchmark.name] = bm_ret.groupby(period_ts).agg('prod') -1
            period_result['Active Return'] = period_result[self.name] - period_result[self.benchmark.name]
            active_risk =  daily_active_ret.groupby(period_ts).agg(lambda s:s.std()*len(s) )
            tolerance = 10**(-5)
            active_risk[active_risk<tolerance] = np.nan
            period_result['Active Risk'] = active_risk
            period_result['Information Ratio'] = period_result['Active Return']/period_result['Active Risk']
            period_result = period_result.drop(index = 0)
            period_result = period_result.style.format('{:.2%}'.format)
            period_result = period_result.format({'Information Ratio': '{:,.2f}'.format})
            self._period_performance = period_result

            return self._period_performance

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




import pandas as pd 

class trading_system:

    def __init__(self, capital=10**6):
        self.capital = 10**6
    
    def set_price(self, OHLC):
        OHLC.columns = list('OHLC')
        OHLC.index = pd.to_datetime(OHLC.index)
        self.__price = OHLC

 ###################   Create Orders   ###############################
    def _init_order_book(self):
        '''
        Initiate the order book to contain order information for execution.
        market orders: record to number of shares to buy/sell in market
        limit orders: a pd.Series with key string format of limit price, value integer format of buy/sell shares.
        target orders: same as limit order, different execution rules.
        '''
        self.market_buy_book = 0
        self.market_sell_book = 0
        self.limit_buy_book = pd.Series()
        self.limit_sell_book = pd.Series()
        self.target_buy_book = pd.Series()
        self.target_sell_book = pd.Series()

    def market_buy(self, quantity):
        '''
        quatity: number of shares to buy at market price.
        '''
        self.market_buy_book += quantity
    
    def market_sell(self, quantity):
        '''
        quantiy: number of shares to sell at market price.
        '''
        self.market_sell_book += quantity

    def limit_buy(self, quantity, price):
        '''
        quantity: number of shares to buy.
        price: highest price to buy
        '''
        self.limit_buy_book[str(price)] = self.limit_buy_book.get(str(price), 0) + quantity

    def limit_sell(self, quantity, price):
        '''
        quantity: number of shares to sell.
        price: lowest price to sell.
        '''
        self.limit_sell_book[str(price)] = self.limit_sell_book.get(str(price), 0) + quantity

    def target_buy(self, quantity, price):
        '''
        quantity: number of shares to buy.
        price: exact price to buy. (No transaction for higher or lower price)
        '''
        self.target_buy_book[str(price)] = self.target_buy_book.get(str(price), 0) + quantity

    def target_sell(self, quantity, price):
        '''
        quantity: number of shares to sell.
        price: exact price to sell. (No transaction for lower or higher price)
        '''
        self.target_sell_book[str(price)] = self.target_buy_book.get(str(price), 0) + quantity

    

#####################   Execution    #########################
    def _init_account(self):
        '''
        On attribute 'account', use a dictionary to store holding information. The dictionary has keys ['cash', 'shares']
        '''
        self.account = {'cash': self.capital, 'share': 0}
        self.transaction = pd.DataFrame(columns=['Date', 'Side', 'Quantity', 'Price'])

    def _load_trading_data(self, date):
        '''
        On attribute 'current_trading_data', use a pd.Series to store OHLC price on given date.
        '''
        date = pd.to_datetime(date)
        self.current_trading_data = self.__price[date, :]
        self.current_trading_date = date

    # to do: handling shorting money/shares? how to deal with remaining orders, return?
    def  transaction(self, date, side, qty, price):
        if side == 'B':
            self.account['cash'] -= qty*price
            self.account['share'] += qty 
            self.transaction.append({'Date':date, 'Side': 'B', 'Quantity':qty, 'Price': price}, ignore_index = True)
        elif side  == 'S':
            self.account['cash'] += qty*price
            self.account['share'] -= qty 
            self.transaction.append({'Date':date, 'Side': 'S', 'Quantity':qty, 'Price': price}, ignore_index = True)
        else:
            raise 'Unknown Side'


    ####################     Market orders     #######################
    def execute_market_orders(self):
        open_price = self.current_trading_data['O']
        date = self.current_trading_date

        # Market trade are execute at open price:
        market_shares = self.market_buy_book -  self.market_sell_book
        if market_shares == 0:
            return
        elif market_shares>0:
            if self.account['cash'] >= market_order_value:
                self.transaction(date, 'B', market_shares, open_price)
                self.market_buy_book = 0
                self.market_sell_book = 0
                return
            else:
                # If not enough capital, then buy as much as possible, non-executed shares will remain:
                trading_share = int(capital/open_price)
                self.transaction(date, 'B', trading_share, open_price)
                self.market_buy_book = market_shares - trading_share
                self.market_sell_book = 0
                return
        else:
            if self.account['share'] >= -market_shares:
                self.transaction(date, 'S', -market_shares, open_price)
                self.market_buy_book = 0
                self.market_sell_book = 0
                return
            else:
                # If not enough shares to sell, then sell all, non-executed shares will remain:
                trading_share = self.account['share']
                self.transaction(date, 'S', trading_share, open_price)
                self.market_buy_book = 0
                self.market_sell_book = -market_shares - trading_share 
                return

    ####################     Target Orders    ####################
    def execute_target_orders(self):
        low_price = self.current_trading_data['L']    
        high_price = self.current_trading_data['H']
        date = self.current_trading_date

        # Target buy:
        numerical_index = pd.to_numeric(self.target_buy_book.index)
        execute_index = (numerical_index <= high_price) & (numerical_index >= low_price)
        for price in numerical_index[execute_index]:
            self.transaction(date, 'B', self.target_buy_book[str(price)], price)
        self.target_buy_book = self.target_buy_book[~execute_index] # Update book

        # Target sell:
        numerical_index = pd.to_numeric(self.target_sell_book.index)
        execute_index = (numerical_index <= high_price) & (numerical_index >= low_price)
        for price in numerical_index[execute_index]:
            self.transaction(date, 'S', self.target_sell_book[str(price)], price)
        self.target_buy_book = self.target_sell_book[~execute_index] # Update book
        
    ####################     Limit Orders   ####################
    def execute_limit_orders(self):
        open_price = self.current_trading_data['O']
        high_price = self.current_trading_data['H']
        low_price = self.current_trading_data['L']
        date = self.current_trading_date

        # open buy:
        numerical_index = pd.to_numeric(self.limit_buy_book.index)
        execute_index = (numerical_index >= open_price)
        for price in numerical_index[execute_index]:
            self.transaction(date, 'B', self.limit_buy_book[str(price)], open_price)
        self.limit_buy_book = self.limit_buy_book[~execute_index] # Update book

        # intraday buy:
        numerical_index = pd.to_numeric(self.limit_buy_book.index)
        execute_index = (numerical_index >= low_price)
        for price in numerical_index[execute_index]:
            self.transaction(date, 'B', self.limit_buy_book[str(price)], price)
        self.limit_buy_book = self.limit_buy_book[~execute_index] # Update book

        # open sell:
        numerical_index = pd.to_numeric(self.limit_sell_book.index)
        execute_index = (numerical_index <= open_price)
        for price in numerical_index[execute_index]:
            self.transaction(date, 'S', self.limit_sell_book[str(price)], open_price)
        self.limit_sell_book = self.limit_sell_book[~execute_index] # Update book

        # intraday sell:
        numerical_index = pd.to_numeric(self.limit_sell_book.index)
        execute_index = (numerical_index <= high_price)
        for price in numerical_index[execute_index]:
            self.transaction(date, 'S', self.limit_sell_book[str(price)], price)
        self.limit_sell_book = self.limit_sell_book[~execute_index] # Update book












class universe:
    def __init__(self, OHLC = dict()):
        '''
        OHLC is a dictionary with key as identity of security (Ticker, SEDOL, ISIN or any other id) and value as data frame with columns ['O', 'H', 'L', 'C'] and rows dates of corresponding data.
        '''
        self.OHLC = OHLC
    
    def append(self, id, OHLC_df):
        self.OHLC[id] = OHLC_df
    
    def merge(self, OHLC_dict):
        self.OHLC = {**self.OHLC, **OHLC_dict}

class orders:
    '''
    Each order should has id (Ticker/SEDOL or anything recognized in the universe), side (B/S), quantity, type (Market/Limit/Target) and price. It should be a pd.Series with corresponding indexes.
    Order record is a data frame of columns ['id', 'side', 'quantity', 'type', 'price']
    '''
    def __init__(self, order_record=None):
        if order_record:
            self.record = order_record
        else:
            self.record = pd.DataFrame(columns=['id', 'side', 'quantity', 'type', 'price'])

    def append(self, order):
        '''
        order is a pd.Series with index ['id', 'side', 'quantity', 'type', 'price'].
        '''
        self.record = self.record.append(order, ignore_index = True)
    
    def merge(self, new_order_record):
        '''
        new_order_record is pd.DataFrame with columns ['id', 'side', 'quantity', 'type', 'price'].
        '''
        self.record = pd.concat([self.record, new_order_record], axis=1, join='inner')

    def delete(self, order=None, order_record=None):
        pass


class account:
    '''
    Account contains information of holding. It is a data frame of columns ['quantity', 'price', 'value'], index as id of security or 'cash'.
    Transactions will record quantity change in account.
    '''
    def __init__(self, capital = 10**6):
        self.holding = pd.DataFrame(columns=['quantity', 'price', 'value'])
        self.holding.loc['Cash', :] = [capital, 1, capital]
    


class signal:

    '''
    signal is a model for evaluating features like lagged OHLC data, fundamental data, related security information, macro data (like GDP growth, interest rate change, FX movement) to output trading recomendation. 
    The API here is much like sk-learn models if it needs to be updated or trained with new data. The model can also be hard-code rules for certain logic. The output can be a direct trade label (Buy, Sell, Hold or numerical values 1, 0, -1) or a rating for each trade label, or a probability distribution of each label.
    '''
    def __init__(self):
        pass







import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="backtest_pkg",
    version="0.1.3",
    author="Andy Hu",
    author_email="andyhu2014@gmail.com",
    description="A package for backtesting portfolio strategies and trading signals",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/andyhu4023/backtest_pkg.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)